#include "pc6001v.h"

#include "common.h"
#include "disk.h"
#include "intr.h"
#include "log.h"
#include "osd.h"
#include "schedule.h"
#include "tape.h"

#include "p6el.h"
#include "p6vm.h"


// P6T`tH[}bgVer.2
//  {Iɂ́ux^C[W+tb^+x^C[WTCY(4byte)vƂ\
//  tb^ɂDATAubN񂪊܂܂B
//  DATAubN͖؂Ƃ1ƐB
//  BASICȂǁuwb_{f[^vƂ\̏ꍇCDATAubN2ƂȂ邪
//  IDԍ𓯈ɂ邱ƂŊ֘AtĊǗłB
//
// [tb^]
//  header (2byte) : "P6"
//  ver    (1byte) : o[W
//  dbnum  (1byte) : ܂܂DATAubN(255܂)
//  start  (1byte) : I[gX^[gtO(0: 1:L)
//  basic  (1byte) : BASIC[h(PC-6001̏ꍇ͖Ӗ)
//  page   (1byte) : y[W
//  askey  (2byte) : I[gX^[gR}h
//  ...R}hꍇ͂̌askey
//  exhead (2byte) : gTCY(64KB܂)
//  ...g񂪂ꍇ͂̌exhead
//
// [DATAubN]
//  header (2byte) : "TI"
//  id     (1byte) : IDԍ(DATAubN֘At)
//  name  (16byte) : f[^(15+'00H')
//  baud   (2byte) : {[[g(600/1200)
//  stime  (2byte) : ̎(ms)
//  ptime  (2byte) : ҁ[̎(ms)
//  offset (4byte) : x^C[W擪̃ItZbg
//  size   (4byte) : f[^TCY


// CxgID
#define	EID_TAPE	(1)

// g
#define PG_HI	(0)
#define PG_LO	(1)

////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
CMTL::CMTL( VM6 *vm, const ID& id ) : Device(vm,id),
	p6t(NULL), Relay(false), stron(false), Boost(DEFAULT_BOOST),
	MaxBoost60(DEFAULT_MAXBOOST60), MaxBoost62(DEFAULT_MAXBOOST62)
{
	INITARRAY( FilePath, '\0' );
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
CMTL::~CMTL( void )
{
	Unmount();
}


////////////////////////////////////////////////////////////////
// CxgR[obN֐
//
// :	id		CxgID
//			clock	NbN
// Ԓl:	Ȃ
////////////////////////////////////////////////////////////////
void CMTL::EventCallback( int id, int clock )
{
	switch( id ){
	case EID_TAPE:	// CMTݏ
		// ݏłȂ?
		if( vm->CpusIsCmtIntrReady() ){
			// CMTAbvf[g
			WORD rd = Update();
			// f[^ȂCMTݔ
			if( rd & PG_D ){
				vm->CpusReqCmtIntr( rd & 0xff );
			}
		}
		break;
	}
}


////////////////////////////////////////////////////////////////
// [g(PLAY,STOP)
////////////////////////////////////////////////////////////////
bool CMTL::Remote( bool relay )
{
    PRINTD( TAPE_LOG, "[TAPE][Relay] -> %s\n", relay ? "true" : "false" );
	
	stron = false;
	
	// [̏Ԃۑ
	Relay = relay;
	
	// [̏ԂɂCxgǉor폜
	if( relay ){	// ON
		// ʃTCYɂBoostUpő{ݒ
		// {BASIC[h(N60,N60m/N66)ɂĔfׂ
		// ܂茵ł͂ȂǂƂɂ
		// SRmk2/66ƓɂĂ݂
		int bst = Boost ? ( vm->VdgGetWinSize() ? MaxBoost60 : MaxBoost62 ) : 1;
		
		if( !vm->EventAdd( this, EID_TAPE, DEFAULT_CMT_HZ * bst, EV_LOOP|EV_HZ ) ) return false;
	}else{			// OFF
		if( !vm->EventDel( this, EID_TAPE ) ) return false;
	}
	return true;
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
bool CMTL::Init( int srate )
{
	// P6T ̈揉(Zbgɂ͖)
	p6t = NULL;
	
	return SndDev::Init( srate );
}


////////////////////////////////////////////////////////////////
// Zbg
////////////////////////////////////////////////////////////////
void CMTL::Reset( void )
{
	// TAPEZbg
	if( p6t ) p6t->Reset();
}


////////////////////////////////////////////////////////////////
// TAPE}Eg
////////////////////////////////////////////////////////////////
bool CMTL::Mount( const char *filename )
{
	PRINTD( TAPE_LOG, "[TAPE][Mount] %s\n", filename  );
	
	// }Egς݂ł΃A}Eg
	if( p6t ){
		delete p6t;
		p6t = NULL;
	}
	
	// P6Tm
	p6t = new cP6T;
	if( !p6t ) return false;
	
	// t@CǍ
	if( !p6t->Readf( filename ) ){
		delete p6t;
		p6t = NULL;
		return false;
	}
	
	// t@CpXۑ
	strncpy( FilePath, filename, PATH_MAX );
	
	return true;
}


////////////////////////////////////////////////////////////////
// TAPEA}Eg
////////////////////////////////////////////////////////////////
void CMTL::Unmount( void )
{
	PRINTD( TAPE_LOG, "[TAPE][Unmount]\n" );
	
	if( p6t ){
		delete p6t;
		p6t = NULL;
		*FilePath = '\0';
	}
}


////////////////////////////////////////////////////////////////
// CMT 1Ǎ
// ߂lFʂ̓f[^(ҁ[,,f[^) ʂ̓f[^
// [ON & CMT OPEN O
////////////////////////////////////////////////////////////////
WORD CMTL::CmtRead( void )
{
	PRINTD( TAPE_LOG, "[TAPE][CmtRead] " );
	
	// TAPE}EgĂȂΖ
	if( !p6t ) return PG_S;
	
	// ҂?
	if( p6t->IsSWaiting() ){
		PRINTD( TAPE_LOG, "swait\n" );
		return PG_S;
	}
	
	// ҁ[҂?
	if( p6t->IsPWaiting() ){
		PRINTD( TAPE_LOG, "pwait\n" );
		return PG_P;
	}
	
	// f[^Ǎ
	PRINTD( TAPE_LOG, "read\n" );
	return PG_D | p6t->ReadOne();
}


////////////////////////////////////////////////////////////////
// }Egς?
////////////////////////////////////////////////////////////////
bool CMTL::IsMount( void ) const
{
	if( p6t ) return true;
	else      return false;
}


////////////////////////////////////////////////////////////////
// I[gX^[g?
////////////////////////////////////////////////////////////////
bool CMTL::IsAutoStart( void ) const
{
	if( p6t ) return p6t->GetAutoStartInfo()->Start;
	else      return false;
}


////////////////////////////////////////////////////////////////
// t@CpX擾
////////////////////////////////////////////////////////////////
const char *CMTL::GetFile() const
{
	return FilePath;
}


////////////////////////////////////////////////////////////////
// TAPE擾
////////////////////////////////////////////////////////////////
const char *CMTL::GetName( void ) const
{
	if( p6t ) return p6t->GetName();
	else      return (const char *)"";
}


////////////////////////////////////////////////////////////////
// x^C[WTCY擾
////////////////////////////////////////////////////////////////
DWORD CMTL::GetSize( void ) const
{
	if( p6t ) return p6t->GetSize();
	else      return 0;
}


////////////////////////////////////////////////////////////////
// JE^擾
////////////////////////////////////////////////////////////////
int CMTL::GetCount( void ) const
{
	if( p6t ) return p6t->GetCount();
	else      return 0;
}


////////////////////////////////////////////////////////////////
// [̏Ԏ擾
////////////////////////////////////////////////////////////////
bool CMTL::IsRelay( void ) const
{
	return Relay;
}


////////////////////////////////////////////////////////////////
// BoostUpݒ
////////////////////////////////////////////////////////////////
void CMTL::SetBoost( bool boost )
{
	if( Boost != boost ){
		Boost = boost;
		// [ONU~߂čĊJ
		if( Relay ){
			Remote( false );
			Remote( true );
		}
	}
}


////////////////////////////////////////////////////////////////
// BoostUpő{ݒ
////////////////////////////////////////////////////////////////
void CMTL::SetMaxBoost( int max60, int max62 )
{
	if( max60 > 0 ) MaxBoost60 = max60;
	if( max62 > 0 ) MaxBoost62 = max62;
}


////////////////////////////////////////////////////////////////
// BoostUpԎ擾
////////////////////////////////////////////////////////////////
bool CMTL::IsBoostUp( void ) const
{
	return Boost;
}


////////////////////////////////////////////////////////////////
// I[gX^[g擾
////////////////////////////////////////////////////////////////
const P6TAUTOINFO *CMTL::GetAutoStartInfo( void ) const
{
	return p6t->GetAutoStartInfo();
}


////////////////////////////////////////////////////////////////
// Xg[XV(1byte)
////////////////////////////////////////////////////////////////
WORD CMTL::Update( void )
{
	PRINTD( TAPE_LOG, "[TAPE][Update]\n" );
	
	WORD rd = PG_S;
	
	// TAPEC[WI[v?
	if( p6t ){
		// 1bytẽf[^ 10ms({)
		int length = SndDev::SampleRate / DEFAULT_CMT_HZ;
		int bdata;
		
		rd = CmtRead();		// CMT 1Ǎ
		
		switch( rd & 0xff00 ){	// f[^`́H
		case PG_P:	// ҁ[̏ꍇ
			// ɃZbg
			while( length-- ) SndDev::cRing::Put( GetSinCurve( PG_HI ) );
			stron = true;	// Xg[XV
			
			break;
		case PG_S:	// ̏ꍇ
			while( length-- ) SndDev::cRing::Put( 0 );
			
			break;
		case PG_D:	// f[^̏ꍇ
			// X^[grbg 1bit
			// f[^rbg   8bits
			// Xgbvrbg 3bits
			
			bdata = length/12;	// 1bit̃f[^
			// X^[grbg
			for( int i=0; i<bdata; i++ ){
				SndDev::cRing::Put( GetSinCurve( PG_LO ) );
			}
			// f[^rbg
			for( int j=7; j>=0; j-- ){
				int hilo = (rd>>j)&1 ? PG_HI : PG_LO;
				for( int i=0; i<bdata; i++ ){
					SndDev::cRing::Put( GetSinCurve( hilo ) );
				}
			}
			// Xgbvrbg
			for( length -= bdata*9; length > 0; length-- ){
				SndDev::cRing::Put( GetSinCurve( PG_HI ) );
			}
			break;
		}
	}
	
	return rd;
}


////////////////////////////////////////////////////////////////
// Xg[XV
//
// :	samples	XVTv(-1:cobt@S 0:NbN)
// Ԓl:	int		XVTv
////////////////////////////////////////////////////////////////
int CMTL::SoundUpdate( int samples )
{
//	PRINTD( TAPE_LOG, "[TAPE][SoundUpdate] Samples: %d(%d)", samples, SndDev::cRing::FreeSize() );
	
	int length = 0;
	
	if( samples == 0 ){
		// Ƃ
	}else if( samples > 0 ) length = min( samples - SndDev::cRing::ReadySize(), SndDev::cRing::FreeSize() );
	else                    length = SndDev::cRing::FreeSize();
	
//	PRINTD( TAPE_LOG, " -> %d\n", length );
	
	if( length <= 0 ) return 0;
	
	for( int i=0; i<length; i++ ){
		// obt@ɏ
		SndDev::cRing::Put( stron ? GetSinCurve( PG_HI ) : 0 );	// 蔲
	}
	
	return length;
}


////////////////////////////////////////////////////////////////
// sing擾
////////////////////////////////////////////////////////////////
int CMTL::GetSinCurve( int fq )
{
	static const int sinc[] = {
		 32767,  32767,  32767,  32767,  32767,  32767,  32767,  32767,  32767,
		 32767,  32767,  32767,  32767,  32767,  32767,  32767,  32767,  32767,
		-32767, -32767, -32767, -32767, -32767, -32767, -32767, -32767, -32767,
		-32767, -32767, -32767, -32767, -32767, -32767, -32767, -32767, -32767
 	};
	static int n=0;
	
	// TvO[gƎgŊԈ̊Ԋu߂
	// f[^e[u44100Hz,1200Hz̐l
	n += (fq == PG_HI ? 2:1 ) * 44100 / SndDev::SampleRate;
	
	// e[uTCY72(sizeof(sinc))
	if( n >= (int)(sizeof(sinc)/sizeof(int)) ) n -= (int)(sizeof(sinc)/sizeof(int));
	
	int ret = ( sinc[n] * SndDev::Volume ) / 100;
	
	return ret;
}




////////////////////////////////////////////////////////////////
// RXgN^
////////////////////////////////////////////////////////////////
CMTS::CMTS( VM6 *vm, const ID& id ) : Device(vm,id), fp(NULL), Baud(1200)
{
	INITARRAY( FilePath, '\0' );
}


////////////////////////////////////////////////////////////////
// fXgN^
////////////////////////////////////////////////////////////////
CMTS::~CMTS( void )
{
	Unmount();
}


////////////////////////////////////////////////////////////////
// 
////////////////////////////////////////////////////////////////
bool CMTS::Init( const char *filename )
{
	if( *filename ){
		// t@CpXۑ
		strncpy( FilePath, filename, PATH_MAX );
	}
	return true;
}


////////////////////////////////////////////////////////////////
// TAPE}Eg
////////////////////////////////////////////////////////////////
bool CMTS::Mount( void )
{
	if( fp ) fclose( fp );
	
	fp = FOPENEN( FilePath, "wb" );
	
	if( fp ) return true;
	else     return false;
}


////////////////////////////////////////////////////////////////
// TAPEA}Eg
////////////////////////////////////////////////////////////////
void CMTS::Unmount( void )
{
	if( fp ){
		// tb^t(Ƃ肠x^)
		fflush( fp );
		int beta = ftell( fp );	// x^C[WTCY擾
		
		// [tb^]
		fputc( 'P', fp );	// header (2byte) : "P6"
		fputc( '6', fp );
		fputc( 2, fp );		// ver    (1byte) : o[W
		fputc( 1, fp );		// dbnum  (1byte) : ܂܂DATAubN(255܂)
		fputc( 0, fp );		// start  (1byte) : I[gX^[gtO(0: 1:L)
		fputc( 1, fp );		// basic  (1byte) : BASIC[h(PC-6001̏ꍇ͖Ӗ)
		fputc( 1, fp );		// page   (1byte) : y[W
		fputc( 0, fp );		// askey  (2byte) : I[gX^[gR}h
		fputc( 0, fp );		// ...R}hꍇ͂̌askey
		fputc( 0, fp );		// exhead (2byte) : gTCY(64KB܂)
		fputc( 0, fp );		// ...g񂪂ꍇ͂̌exhead
		
		// [DATAubN]
		fputc( 'T', fp );				// header (2byte) : "TI"
		fputc( 'I', fp );
		fputc( 0, fp );					// id     (1byte) : IDԍ(DATAubN֘At)
		for( int i=0; i<16; i++ ) fputc( 0, fp );	// name  (16byte) : f[^(15+'00H')
		FPUTWORD( Baud, fp );			// baud   (2byte) : {[[g(600/1200)
		fputc( 0x48, fp );				// stime  (2byte) : ̎(ms)
		fputc( 0x0d, fp );
		fputc( 0x48, fp );				// ptime  (2byte) : ҁ[̎(ms)
		fputc( 0x0d, fp );
		fputc( 0, fp );					// offset (4byte) : x^C[W擪̃ItZbg
		fputc( 0, fp );
		fputc( 0, fp );
		fputc( 0, fp );
		FPUTDWORD( beta, fp );			// size   (4byte) : f[^TCY
		
		// [x^C[WTCY]
		FPUTDWORD( beta, fp );
		
		fclose( fp );
		fp = NULL;
		*FilePath = '\0';
	}
}


////////////////////////////////////////////////////////////////
// {[[gݒ
////////////////////////////////////////////////////////////////
void CMTS::SetBaud( int b )
{
	Baud = b;
}


////////////////////////////////////////////////////////////////
// CMT 1
////////////////////////////////////////////////////////////////
void CMTS::CmtWrite( BYTE data )
{
	if( fp ) fputc( data, fp );
}


////////////////////////////////////////////////////////////////
// I/OANZX֐
////////////////////////////////////////////////////////////////
void CMTL::OutB0H( int, BYTE data ){ Remote( data&0x08 ? true : false ); }


////////////////////////////////////////////////////////////////
// ǂłSAVE
////////////////////////////////////////////////////////////////
bool CMTL::DokoSave( cIni *Ini )
{
	if( !Ini ) return false;
	
	Ini->PutEntry( "TAPE", NULL, "Relay",	"%s",	Relay ? "Yes" : "No" );
	Ini->PutEntry( "TAPE", NULL, "BoostUp",	"%s",	Boost ? "Yes" : "No" );
	
	// TAPE}EgĂȂΉȂŖ߂
	if( !p6t ) return true;
	
	// }EgĂP6TIuWFNgSAVE
	char pathstr[PATH_MAX+1];
	strncpy( pathstr, FilePath, PATH_MAX );
	OSD_RelativePath( pathstr );
	Ini->PutEntry( "TAPE", NULL, "FilePath",	"%s",	pathstr );
	
	return p6t->DokoSave( Ini );
}


////////////////////////////////////////////////////////////////
// ǂłLOAD
////////////////////////////////////////////////////////////////
bool CMTL::DokoLoad( cIni *Ini )
{
	if( !Ini ) return false;
	
	Ini->GetTruth( "TAPE", "Relay",		&Relay,	Relay );
	Ini->GetTruth( "TAPE", "BoostUp",	&Boost,	Boost );
	
	Ini->GetString( "TAPE", "FilePath", FilePath, "" );
	if( *FilePath ){
		if( !Mount( FilePath ) ) return false;
		if( !p6t->DokoLoad( Ini ) ) return false;
	}else
		Unmount();
	
	
	return true;
}


////////////////////////////////////////////////////////////////
//  device description
////////////////////////////////////////////////////////////////
const Device::Descriptor CMTL::descriptor = {
	CMTL::indef, CMTL::outdef
};

const Device::OutFuncPtr CMTL::outdef[] = {
	STATIC_CAST( Device::OutFuncPtr, &CMTL::OutB0H )
};

const Device::InFuncPtr CMTL::indef[] = { NULL };
